@@ -59,6 +59,7 @@ static const char *vdagentd_socket = VDAGENTD_SOCKET;
 static const char *uinput_device = "/dev/uinput";
 static int debug = 0;
 static int uinput_fake = 0;
+static int only_once = 0;
 static struct udscs_server *server = NULL;
 static struct vdagent_virtio_port *virtio_port = NULL;
 static GHashTable *active_xfers = NULL;
@@ -74,6 +75,7 @@ static int agent_owns_clipboard[256] = { 0, };
 static int quit = 0;
 static int retval = 0;
 static int client_connected = 0;
+static int max_clipboard = -1;
 
 /* utility functions */
 /* vdagentd <-> spice-client communication handling */
@@ -98,6 +100,8 @@ static void send_capabilities(struct vdagent_virtio_port *vport,
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_CLIPBOARD_SELECTION);
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_SPARSE_MONITORS_CONFIG);
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_GUEST_LINEEND_LF);
+    VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_MAX_CLIPBOARD);
+    VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_AUDIO_VOLUME_SYNC);
 
     vdagent_virtio_port_write(vport, VDP_CLIENT_PORT,
                               VD_AGENT_ANNOUNCE_CAPABILITIES, 0,
@@ -153,6 +157,19 @@ static void do_client_monitors(struct vdagent_virtio_port *vport, int port_nr,
                               (uint8_t *)&reply, sizeof(reply));
 }
 
+static void do_client_volume_sync(struct vdagent_virtio_port *vport, int port_nr,
+    VDAgentMessage *message_header,
+    VDAgentAudioVolumeSync *avs)
+{
+    if (active_session_conn == NULL) {
+        syslog(LOG_DEBUG, "No active session - Can't volume-sync");
+        return;
+    }
+
+    udscs_write(active_session_conn, VDAGENTD_AUDIO_VOLUME_SYNC, 0, 0,
+                (uint8_t *)avs, message_header->size);
+}
+
 static void do_client_capabilities(struct vdagent_virtio_port *vport,
     VDAgentMessage *message_header,
     VDAgentAnnounceCapabilities *caps)
@@ -369,6 +386,20 @@ int virtio_port_read_complete(
         vdagent_virtio_port_reset(vport, VDP_CLIENT_PORT);
         do_client_disconnect();
         break;
+    case VD_AGENT_MAX_CLIPBOARD:
+        if (message_header->size != sizeof(VDAgentMaxClipboard))
+            goto size_error;
+        VDAgentMaxClipboard *msg = (VDAgentMaxClipboard *)data;
+        syslog(LOG_DEBUG, "Set max clipboard: %d", msg->max);
+        max_clipboard = msg->max;
+        break;
+    case VD_AGENT_AUDIO_VOLUME_SYNC:
+        if (message_header->size < sizeof(VDAgentAudioVolumeSync))
+            goto size_error;
+
+        do_client_volume_sync(vport, port_nr, message_header,
+                (VDAgentAudioVolumeSync *)data);
+        break;
     default:
         syslog(LOG_WARNING, "unknown message type %d, ignoring",
                message_header->type);
@@ -382,6 +413,34 @@ size_error:
     return 0;
 }
 
+static void virtio_write_clipboard(uint8_t selection, uint32_t msg_type,
+    uint32_t data_type, const uint8_t *data, uint32_t data_size)
+{
+    uint32_t size = data_size;
+
+    if (VD_AGENT_HAS_CAPABILITY(capabilities, capabilities_size,
+                                VD_AGENT_CAP_CLIPBOARD_SELECTION)) {
+        size += 4;
+    }
+    if (data_type != -1) {
+        size += 4;
+    }
+
+    vdagent_virtio_port_write_start(virtio_port, VDP_CLIENT_PORT, msg_type,
+                                    0, size);
+
+    if (VD_AGENT_HAS_CAPABILITY(capabilities, capabilities_size,
+                                VD_AGENT_CAP_CLIPBOARD_SELECTION)) {
+        uint8_t sel[4] = { selection, 0, 0, 0 };
+        vdagent_virtio_port_write_append(virtio_port, sel, 4);
+    }
+    if (data_type != -1) {
+        vdagent_virtio_port_write_append(virtio_port, (uint8_t*)&data_type, 4);
+    }
+
+    vdagent_virtio_port_write_append(virtio_port, data, data_size);
+}
+
 /* vdagentd <-> vdagent communication handling */
 int do_agent_clipboard(struct udscs_connection *conn,
         struct udscs_message_header *header, const uint8_t *data)
@@ -425,12 +484,21 @@ int do_agent_clipboard(struct udscs_connection *conn,
     case VDAGENTD_CLIPBOARD_DATA:
         msg_type = VD_AGENT_CLIPBOARD;
         data_type = header->arg2;
+        if (max_clipboard != -1 && size > max_clipboard) {
+            syslog(LOG_WARNING, "clipboard is too large (%d > %d), discarding",
+                   size, max_clipboard);
+            virtio_write_clipboard(selection, msg_type, data_type, NULL, 0);
+            return 0;
+        }
         break;
     case VDAGENTD_CLIPBOARD_RELEASE:
         msg_type = VD_AGENT_CLIPBOARD_RELEASE;
         size = 0;
         agent_owns_clipboard[selection] = 0;
         break;
+    default:
+        syslog(LOG_WARNING, "unexpected clipboard message type");
+        goto error;
     }
 
     if (size != header->size) {
@@ -439,27 +507,7 @@ int do_agent_clipboard(struct udscs_connection *conn,
         return -1;
     }
 
-    if (VD_AGENT_HAS_CAPABILITY(capabilities, capabilities_size,
-                                VD_AGENT_CAP_CLIPBOARD_SELECTION)) {
-        size += 4;
-    }
-    if (data_type != -1) {
-        size += 4;
-    }
-
-    vdagent_virtio_port_write_start(virtio_port, VDP_CLIENT_PORT, msg_type,
-                                    0, size);
-
-    if (VD_AGENT_HAS_CAPABILITY(capabilities, capabilities_size,
-                                VD_AGENT_CAP_CLIPBOARD_SELECTION)) {
-        uint8_t sel[4] = { selection, 0, 0, 0 };
-        vdagent_virtio_port_write_append(virtio_port, sel, 4);
-    }
-    if (data_type != -1) {
-        vdagent_virtio_port_write_append(virtio_port, (uint8_t*)&data_type, 4);
-    }
-
-    vdagent_virtio_port_write_append(virtio_port, data, header->size);
+    virtio_write_clipboard(selection, msg_type, data_type, data, header->size);
 
     return 0;
 
@@ -730,7 +778,9 @@ static void usage(FILE *fp)
             "  -s <port>      set virtio serial port  [%s]\n"
             "  -S <filename>  set udcs socket [%s]\n"
             "  -u <dev>       set uinput device       [%s]\n"
+            "  -f             treat uinput device as fake; no ioctls\n"
             "  -x             don't daemonize\n"
+            "  -o             Only handle one virtio serial session.\n"
 #ifdef HAVE_CONSOLE_KIT
             "  -X             Disable console kit integration\n"
 #endif
@@ -771,6 +821,7 @@ void main_loop(void)
     fd_set readfds, writefds;
     int n, nfds;
     int ck_fd = 0;
+    int once = 0;
 
     while (!quit) {
         FD_ZERO(&readfds);
@@ -800,6 +851,7 @@ void main_loop(void)
         udscs_server_handle_fds(server, &readfds, &writefds);
 
         if (virtio_port) {
+            once = 1;
             vdagent_virtio_port_handle_fds(&virtio_port, &readfds, &writefds);
             if (!virtio_port) {
                 int old_client_connected = client_connected;
@@ -818,6 +870,11 @@ void main_loop(void)
                 client_connected = old_client_connected;
             }
         }
+        else if (only_once && once)
+        {
+            syslog(LOG_INFO, "Exiting after one client session.");
+            break;
+        }
 
         if (session_info && FD_ISSET(ck_fd, &readfds)) {
             active_session = session_info_get_active_session(session_info);
@@ -839,7 +896,7 @@ int main(int argc, char *argv[])
     struct sigaction act;
 
     for (;;) {
-        if (-1 == (c = getopt(argc, argv, "-dhxXs:u:S:")))
+        if (-1 == (c = getopt(argc, argv, "-dhxXfos:u:S:")))
             break;
         switch (c) {
         case 'd':
@@ -854,6 +911,12 @@ int main(int argc, char *argv[])
         case 'u':
             uinput_device = optarg;
             break;
+        case 'f':
+            uinput_fake = 1;
+            break;
+        case 'o':
+            only_once = 1;
+            break;
         case 'x':
             do_daemonize = 0;
             break;
@@ -870,11 +933,6 @@ int main(int argc, char *argv[])
         }
     }
 
-    if (strncmp(uinput_device, "/dev", 4) != 0) {
-        syslog(LOG_INFO, "using fake uinput");
-        uinput_fake = 1;
-    }
-
     memset(&act, 0, sizeof(act));
     act.sa_flags = SA_RESTART;
     act.sa_handler = quit_handler;
